import { walk } from 'estree-walker';
// Transformers
import transformStyles from './styles.js';
import transformDoctype from './doctype.js';
import transformModuleScripts from './module-scripts.js';
import transformCodeBlocks from './prism.js';
/** Add visitors to given collection */
function addVisitor(visitor, collection, nodeName, event) {
    if (typeof visitor[event] !== 'function')
        return;
    if (!collection[event])
        collection[event] = new Map();
    const visitors = collection[event].get(nodeName) || [];
    visitors.push(visitor[event]);
    collection[event].set(nodeName, visitors);
}
/** Compile visitor actions from transformer */
function collectVisitors(transformer, htmlVisitors, cssVisitors, finalizers) {
    if (transformer.visitors) {
        if (transformer.visitors.html) {
            for (const [nodeName, visitor] of Object.entries(transformer.visitors.html)) {
                addVisitor(visitor, htmlVisitors, nodeName, 'enter');
                addVisitor(visitor, htmlVisitors, nodeName, 'leave');
            }
        }
        if (transformer.visitors.css) {
            for (const [nodeName, visitor] of Object.entries(transformer.visitors.css)) {
                addVisitor(visitor, cssVisitors, nodeName, 'enter');
                addVisitor(visitor, cssVisitors, nodeName, 'leave');
            }
        }
    }
    finalizers.push(transformer.finalize);
}
/** Utility for formatting visitors */
function createVisitorCollection() {
    return {
        enter: new Map(),
        leave: new Map(),
    };
}
/** Walk AST with collected visitors */
function walkAstWithVisitors(tmpl, collection) {
    walk(tmpl, {
        enter(node, parent, key, index) {
            if (collection.enter.has(node.type)) {
                const fns = collection.enter.get(node.type);
                for (let fn of fns) {
                    fn.call(this, node, parent, key, index);
                }
            }
        },
        leave(node, parent, key, index) {
            if (collection.leave.has(node.type)) {
                const fns = collection.leave.get(node.type);
                for (let fn of fns) {
                    fn.call(this, node, parent, key, index);
                }
            }
        },
    });
}
/**
 * Transform
 * Step 2/3 in Astro SSR.
 * Transform is the point at which we mutate the AST before sending off to
 * Codegen, and then to Snowpack. In some ways, it‘s a preprocessor.
 */
export async function transform(ast, opts) {
    const htmlVisitors = createVisitorCollection();
    const cssVisitors = createVisitorCollection();
    const finalizers = [];
    const optimizers = [transformStyles(opts), transformDoctype(opts), transformModuleScripts(opts), transformCodeBlocks(ast.module)];
    for (const optimizer of optimizers) {
        collectVisitors(optimizer, htmlVisitors, cssVisitors, finalizers);
    }
    walkAstWithVisitors(ast.css, cssVisitors);
    walkAstWithVisitors(ast.html, htmlVisitors);
    // Run all of the finalizer functions in parallel because why not.
    await Promise.all(finalizers.map((fn) => fn()));
}
